from ..language_object import CoreObjectInterface, LanguageObjectInterface
from ..error import LineupError
from typing import Dict, Any, List, Tuple


class VariableNotExistError(LineupError):
    pass


class VariableNotLanguageObjectError(LineupError):
    pass


class Variables(CoreObjectInterface):
    variables: Dict[str, Any]
    default_variables: List[str]
    version = None

    def __init__(self, variables: Dict[str, Any] | None) -> None:
        if variables is None:
            variables = {}
        self.variables = variables
        self.default_variables = list(variables.keys())
        self.default_values = variables.copy()
        self.functions = {
            "VAR": self._variable,
        }

    def close(self):
        super().close()
        for key in list(self.variables.keys()):
            if isinstance(self.variables[key], LanguageObjectInterface):
                self.variables[key].close()
            self.variables.pop(key)

    def reset(self):
        error = []
        error += self._reset_disregard()
        error += self._reset_default()
        if error:
            for key, e in error:
                self.logger.error(f"Error at close with {key}: {e}")
            raise LineupError(f"Error with {len(error)} variables")

    def _reset_disregard(self) -> List[Tuple[str, Exception]]:
        variables_to_delete = [key for key in self.variables.keys()
                               if key not in self.default_variables]
        error = []
        for key in variables_to_delete:
            if isinstance(self.variables[key], LanguageObjectInterface):
                self.logger.debug(f"{self} close {key}")
                try:
                    self.variables[key].close()
                except Exception as e:
                    error.append((key, e))
            self.variables.pop(key)
        return error

    def _reset_default(self) -> List[Tuple[str, Exception]]:
        error = []
        for key in self.variables.keys():
            if isinstance(self.variables[key], LanguageObjectInterface):
                self.logger.debug(f"{self} reset {key}")
                try:
                    self.variables[key].reset()
                except Exception as e:
                    error.append((key, e))
            else:
                self.variables[key] = self.default_values[key]
        return error

    def _get(self, name: str):
        if name in self.variables:
            return self.variables[name]
        msg = f"'{name}' not exist in '{self}'"
        self.logger.error(msg)
        raise VariableNotExistError(msg)

    def _set(self, name: str, value):
        self.variables[name] = value

    def _delete(self, name: str):
        if name not in self.variables:
            raise VariableNotExistError(f"'{name}' not exist in '{self}'")
        del self.variables[name]

    def _execute_in_variables(self, variables, function_name: str, *args):
        if variables not in self.variables:
            raise VariableNotExistError(f"'{variables}' not exist in '{self}'")
        if not isinstance(self.variables[variables], LanguageObjectInterface):
            raise VariableNotLanguageObjectError(f"'{variables}' is not a LanguageObjectInterface in '{self}'")
        return self.variables[variables].execute(function_name, *args)

    def _execute_from_executor(self, line: List[str]):
        return self.executor.execute_line(line)

    def _exit(self, *args):
        self.executor.stop = True
        return self._execute_from_executor(args)

    def _variable(self, name: str, command: str, *args):
        match command:
            case "GET":
                return self._get(name)
            case "SET":
                self._set(name, args[0])
            case "USE":
                return self._set(name, self._execute_from_executor(args))
            case "UNSET":
                self._delete(name)
            case "EXEC":
                return self._execute_in_variables(name, *args)
            case _:
                return self._execute_in_variables(name, command, *args)
        return None



from .default import DefaultExecutor
from ..error import ExecutorFunctionNotExistError, LineupError
from typing import Any, List


class JumperOffsetInvalidError(LineupError):
    pass


class JumperExecutor(DefaultExecutor):
    jump_functions = ["JUMP", "GOTO"]
    line = 0
    # Executor version - it's set on the lineup version
    _version = None

    def jump(self, offset: int):
        if offset == 0:
            raise JumperOffsetInvalidError("The offset can't be 0")
        self.line += offset - 1
        self.logger.debug(f"Jump to line: {self.line + 1}")

    def goto(self, line: int):
        """
        Execute the line passed as argument
        """
        if line <= 0:
            raise JumperOffsetInvalidError("The line can't be 0")
        # -2 because the line is 1-based and the list is 0-based
        # and -1 because the line we want will be the next one
        self.line = line - 2
        self.logger.debug(f"Go to line: {self.line + 1}")

    def execute_jump(self, line: List[str]):
        if len(line) != 2:
            raise ExecutorFunctionNotExistError(
                f"'{line[0]}' need 1 argument, got {len(line) - 1}")
        match line[0]:
            case "JUMP":
                return self.jump(int(line[1]))
            case "GOTO":
                return self.goto(int(line[1]))
        raise ExecutorFunctionNotExistError(
            f"'{line[0]}' not exist in '{self}'")

    def execute_line(self, line: List[str]):
        if line[0] in self.jump_functions:
            self.logger.debug(f"Execute: {line}")
            return self.execute_jump(line)
        return super().execute_line(line)

    def execute(self, script: List[List[str]]) -> Any:
        self.stop = False
        result = None
        self.line = 0
        while self.line < len(script):
            if self.stopped:
                break
            result = self.execute_line(script[self.line])
            self.line += 1
        return result



from .error import DecodeLineStringError
from typing import List
import regex as re
import json


class LineDecoder:
    """
    Decode a line string to an instruction list
    """

    def decode(self, line: str) -> List[str] | None:
        """
        Decode a line string to an instruction list
        """
        line = line.strip()
        if line.startswith("#"):
            return None
        # delete comments on the line
        line = self._delete_comments(line)
        line_splitted = self._split_line(line)
        if not line_splitted:
            return None
        return line_splitted

    def _delete_comments(self, line: str) -> str:
        """
        Delete the comments on the line
        """
        lines = line.split("#")
        result = []
        for i in lines:
            i, _ = self._format_escape(i, False, "#")
            result.append(i)
            if len(i) == 0 or i[-1] != "#":
                break
        return "".join(result)

    def _split_line(self, line: str) -> List[str]:
        lines = self.format_quotes(line)
        result = []
        if len(lines) % 2 == 0:
            raise DecodeLineStringError(f"'{line}' is not valid line string")
        # odd is inside quotes, even is outside
        for i, line in enumerate(lines):
            if i % 2 == 1:
                # inside quotes, the line has to be kept as is
                try:
                    result.append(json.loads(f'"{line}"'))
                except json.JSONDecodeError:
                    raise DecodeLineStringError(f"'{line}' is not valid line string")
            else:
                # outside quotes, split by space for get the instructions
                result.extend([
                    x for x in line.split(" ") if x.strip() != ""
                ])
        return result

    def format_quotes(self, line: str) -> List[str]:
        """
        Format the quotes in the line

        Split the line by quotes, keep the backslahed quotes
        Return a list of odd is inside quotes, even is outside the quotes
        """
        lines = line.split('"')
        result = []
        current_line = ""
        if len(lines) == 1:
            return [line]
        for i in lines:
            # if len(i) % 2 != 1 it means that the line is inside quotes
            # because we don't already write on odd indexes
            i, is_ended = self._format_escape(i, len(result) % 2 == 1)
            current_line += i
            if is_ended:
                result.append(current_line)
                current_line = ""
        if current_line:
            raise DecodeLineStringError(f"'{line}' is not valid line string")
        return result

    def _format_escape(self, line: str, inside_quotes: bool,
                       ended_char: str = '"') -> tuple[str, bool]:
        """
        Format the escape characters in the line"
        """
        if not line:
            return line, True
        nb_backslash_at_end = self.count_backslash_at_end(line)
        if nb_backslash_at_end == 0:
            return line, True
        if not inside_quotes:
            # Outside a quote, we have to remove the backslashes ourselves
            line = line[:-nb_backslash_at_end] + ("\\" * (nb_backslash_at_end // 2))
        # Inside a quote, it's JSON, so we have to keep the backslashes
        if nb_backslash_at_end % 2 == 0:
            return line, True
        line = line + ended_char
        return line, False

    def count_backslash_at_end(self, line: str) -> int:
        """
        Count the number of backslashes at the end of the line
        """
        nb_backslash = 0
        for i in range(len(line) - 1, -1, -1):
            if line[i] == "\\":
                nb_backslash += 1
            else:
                break
        return nb_backslash



from __future__ import annotations
from typing import Dict, Any, Callable, List
from .error import FunctionNotExistError, AlreadyClosedError
import logging


class LineupObjectInterface:
    """
    LineUp Object, who can be used by the language
    Can be close or reset
    """

    _is_closed = False
    """
    If the object is closed
    If it's closed, the object can't be used anymore
    """

    def close(self):
        """
        Destroy the object
        """
        self._is_closed = True

    def reset(self):
        """
        Reset the object for future new exectution
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")


class LanguageObjectInterface(LineupObjectInterface):
    """
    LineUp Object with functions who can be executed by the language
    """

    functions: Dict[str, Callable[..., Any]]
    """
    All functions the script can execute
    """

    logger = logging.getLogger("lineup_lang")

    def get_all_functions(self) -> List[str]:
        """
        Get all functions name in the object
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")
        return list(self.functions.keys())

    def execute(self, function_name: str, *args) -> Any:
        """
        Execute a function in the object
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")
        if function_name not in self.functions:
            msg = f"'{function_name}' not exist in '{self}'"
            self.logger.error(msg)
            raise FunctionNotExistError(msg)
        return self.functions[function_name](*args)

    def __str__(self) -> str:
        return f"<LUPO:{self.__class__.__name__}>"


class CoreObjectInterface(LanguageObjectInterface):
    """
    Object instancied before the creation of the language executor

    Function in this object is directly accessible by the language executor

    A core object is used for create the main function of the language executor
    """
    executor: LanguageExecutorInterface
    version: str = "0.0.0"

    def set_executor(self, executor: LanguageExecutorInterface) -> None:
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")
        self.executor = executor

    def get_version(self) -> str:
        return self.version

    def __str__(self) -> str:
        return f"<LUPC:{self.__class__.__name__}>"


class LanguageExecutorInterface(LineupObjectInterface):
    """
    Language executor interface

    This object is the main object who execute the script
    It execute line each by each
    """
    _core_function: Dict[str, LanguageObjectInterface]
    _core: List[LanguageObjectInterface]
    _version: str = "0.0.0"
    stopped: bool = False

    def execute_line(self, line: List[str]):
        """
        Execute one line of the script
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")

    def execute(self, script: List[List[str]]) -> Any:
        """
        Execute a list of line
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")

    def get_versions(self) -> Dict[str, str]:
        """
        Get all the versions who defined the language
        (executor and core object)
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")
        versions = {self.__str__(): self._version}
        for core in self._core:
            versions[core.__str__()] = core.get_version()
        return versions

    def close(self) -> None:
        """
        Close the executor

        This function is called when the executor is not used anymore
        """
        if self._is_closed:
            return
        super().close()
        for core in self._core:
            logging.getLogger("lineup_lang").info(f"Close: {core}")
            core.close()

    def reset(self) -> None:
        """
        Reset the executor for future new exectution

        This function is called when after a script execution
        """
        if self._is_closed:
            raise AlreadyClosedError(f"{self} is already closed")
        for core in self._core:
            logging.getLogger("lineup_lang").info(f"Reset: {core}")
            core.reset()

    def __str__(self) -> str:
        return f"<LUPE:{self.__class__.__name__}>"


class LanguageInterface(LineupObjectInterface):
    """"
    Language interface
    It read the script, cut it in line and
    send it to the executor for execute it
    """

    def get_all_functions(self) -> List[str]:
        """
        Get all functions name in the language
        """
        pass

    def get_versions(self) -> str:
        """
        Get all the versions who defined the language

        (executor, core object and lineup version)
        """
        pass

    def execute_script(self, script: str) -> Any:
        """
        Execute a script
        """
        pass

    def execute_script_with_args(self, script: str, **kwargs) -> Any:
        """
        Execute a script with arguments

        Arguments in the script is write like this:
        $LUP_ARG_NAME (in the script) -> kwargs["LUP_ARG_NAME"] (in the function)
        ${LUP_ARG_NAME} (in the script) -> kwargs["LUP_ARG_NAME"] (in the function)
        """
        pass

    def execute_file(self, file_path: str, **kwargs) -> Any:
        """
        Execute a file with arguments

        Arguments in the script is write like this:
        $LUP_ARG_NAME (in the script) -> kwargs["LUP_ARG_NAME"] (in the function)
        ${LUP_ARG_NAME} (in the script) -> kwargs["LUP_ARG_NAME"] (in the function)
        """
        pass

    def __str__(self) -> str:
        return f"<LUPL:{self.__class__.__name__}>"



from ..language_object import CoreObjectInterface
from ..line_decoder import LineDecoder
from ..error import LineupError


class Conditions(CoreObjectInterface):
    last_result = None
    version = None

    equal_signs = {
        "EQ": "==",
        "NE": "!=",
        "GT": ">",
        "GE": ">=",
        "LT": "<",
        "LE": "<=",
        "==": "==",
        "!=": "!=",
        ">": ">",
        ">=": ">=",
        "<": "<",
        "<=": "<=",
    }

    def __init__(self):
        self.functions = {
            "IF": self._if,
            "NOTIF": self._notif,
            "ELSE": self._else,
        }

    def _execute_jump(self, jump_size: str):
        if jump_size.startswith("*"):
            self.executor.execute_line(["JUMP", jump_size[1:]])
        else:
            self.executor.execute_line(["GOTO", jump_size])

    def _if(self, jump_size: str, *args):
        result = self._execute_condition(*args)
        if result:
            self.last_result = True
            self._execute_jump(jump_size)
        self.last_result = False
        return result

    def _notif(self, jump_size: str, *args):
        result = self._execute_condition(*args)
        if not result:
            self.last_result = True
            self._execute_jump(jump_size)
        self.last_result = False
        return result

    def _else(self, *args):
        if self.last_result is None:
            raise LineupError("ELSE without IF")
        if not self.last_result:
            self._execute_jump(args[0])
        self.last_result = False

    def _execute_condition(self, *args):
        if len(args) == 3 and args[1] in self.equal_signs.keys():
            return self._equal(args[0], args[1], args[2])
        return self.executor.execute_line(args)

    def _equal(self, arg1: str, sign: str, arg2: str) -> bool:
        sign = self.equal_signs[sign]
        line_decoder = LineDecoder()
        arg1 = line_decoder.decode(arg1)
        arg2 = line_decoder.decode(arg2)
        result_1 = self.executor.execute_line(arg1)
        result_2 = self.executor.execute_line(arg2)
        match sign:
            case "==":
                return result_1 == result_2
            case "!=":
                return result_1 != result_2
            case ">":
                return result_1 > result_2
            case ">=":
                return result_1 >= result_2
            case "<":
                return result_1 < result_2
            case "<=":
                return result_1 <= result_2
        return False



import logging


def start_logging(log_level: str = "WARN") -> None:
    """
    Add handlers to the logger if not exist

    :param log_level: Log level
    """
    if logging.getLogger("lineup_lang").hasHandlers():
        return
    logger = logging.getLogger("lineup_lang")
    logger.setLevel(log_level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)



from typing import List, Any, Dict
from .language_object import LanguageInterface, LanguageExecutorInterface, \
     LanguageObjectInterface, CoreObjectInterface
from .error import LineupError, UnexpectedError
from .logger import start_logging
from .line_decoder import LineDecoder
from .args_resolver import ArgsResolver
import lineup_lang.executor as luexec
import lineup_lang.core as lucore
import logging
import os

__all__ = ["Language", "LanguageObjectInterface", "CoreObjectInterface", "luexec", "lucore"]


class Language(LanguageInterface):
    """Language object"""

    _executor: LanguageExecutorInterface
    no_error: bool
    logger = logging.getLogger("lineup_lang")
    is_closed = False

    def __init__(self, executor: LanguageExecutorInterface,
                 no_error: bool = False, log_level: str = "WARN"):
        start_logging(log_level)
        self._executor = executor
        self.no_error = no_error
        self.logger.info(f"Start: {self} with executor: {executor}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        if self.is_closed:
            return
        self.is_closed = True
        self.logger.info(f"Close: {self}")
        self._executor.close()

    def reset(self):
        super().reset()
        self._executor.reset()

    def get_all_functions(self) -> List[str]:
        return self._executor.get_all_functions()

    def get_versions(self) -> Dict[str, str]:
        file_version = os.path.join(os.path.dirname(__file__), "VERSION")
        lup_version = "0.0.0"
        with open(file_version, "r") as file:
            lup_version = file.read().strip()
        all_versions = self._executor.get_versions()
        all_versions["lineup_lang"] = lup_version
        for [name, version] in all_versions.items():
            if version is None:
                all_versions[name] = lup_version
        return all_versions

    def execute_script(self, script: str) -> Any:
        self.logger.info("Launch a script")
        self.logger.debug(f"Execute script:\n{script}")
        line_decoder = LineDecoder()
        script_lines = []
        for line in script.split("\n"):
            line = line_decoder.decode(line)
            if not line:
                continue
            script_lines.append(line)
        try:
            if self.is_closed:
                raise LineupError("Language is closed")
            result = self._executor.execute(script_lines)
        except Exception as e:
            if self.no_error:
                self.logger.error(f"Error: {e.__class__.__name__}: {e}")
                return None
            if isinstance(e, LineupError):
                raise e
            raise UnexpectedError(e)
        self._executor.reset()
        return result

    def execute_script_with_args(self, script: str, **kwargs) -> Any:
        args_resolver = ArgsResolver()
        script = args_resolver.resolve(script, **kwargs)
        return self.execute_script(script)

    def execute_file(self, file_path: str, **kwargs) -> Any:
        with open(file_path, "r") as file:
            script = file.read()
        return self.execute_script_with_args(script, **kwargs)

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}>"



class LineupError(Exception):
    """Base class for exceptions in this module."""
    pass


class ExecutorFunctionAlreadyExistError(LineupError):
    """Two core object define the same function"""
    pass


class FunctionNotExistError(LineupError):
    """Function not exist in a language object"""
    pass


class ExecutorFunctionNotExistError(FunctionNotExistError):
    """Function not exist in a language executor"""
    pass


class ArgumentNotExistError(LineupError):
    """The script need an argument that is not present"""
    pass


class DecodeLineStringError(LineupError):
    """An error occur when decoding a line"""
    pass


class AlreadyClosedError(LineupError):
    """An object is already closed"""
    pass


class UnexpectedError(LineupError):
    """An unexpected error occur, who are not managed by the language"""

    def __init__(self, error: Exception):
        super().__init__(f"An unexpected error occur: {error.__class__.__name__}: {error}")



from ..language_object import LanguageExecutorInterface, \
    CoreObjectInterface
from ..error import ExecutorFunctionAlreadyExistError, \
    ExecutorFunctionNotExistError
from typing import Any, List
import logging


class DefaultExecutor(LanguageExecutorInterface):
    logger = None
    # Executor version - it's set on the lineup version
    _version = None

    def __init__(self, core_object: List[CoreObjectInterface]):
        self.logger = logging.getLogger("lineup_lang")
        self._core_function = {}
        self._core = []
        for core in core_object:
            self._core.append(core)
            core.set_executor(self)
            for function_name in core.get_all_functions():
                if function_name in self._core_function:
                    fn = function_name
                    c1 = core
                    c2 = self._core_function[function_name]
                    raise ExecutorFunctionAlreadyExistError(
                        f"'{fn}' from '{c1}' in '{c2}'")
                self._core_function[function_name] = core

    def reset(self) -> None:
        self.stopped = False
        super().reset()

    def execute_line(self, line: List[str]):
        super().execute_line(line)
        if line[0] not in self._core_function:
            msg = f"'{line[0]}' not exist in '{self}'"
            self.logger.error(msg)
            raise ExecutorFunctionNotExistError(msg)
        self.logger.debug(f"Execute: {line}")
        return self._core_function[line[0]].execute(line[0], *line[1:])

    def execute(self, script: List[List[str]]) -> Any:
        super().execute(script)
        self.stopped = False
        result = None
        for line in script:
            result = self.execute_line(line)
            if self.stopped:
                break
        return result



from ..language_object import CoreObjectInterface


class System(CoreObjectInterface):
    def __init__(self):
        self.functions = {
            "EXIT": self._exit,
            "DEBUG": self._debug,
            "OBJ": self._obj,
        }

    def _exit(self, *args):
        result = None
        if len(args) > 0:
            result = self.executor.execute_line(args)
        self.executor.stopped = True
        return result

    def _debug(self, *args):
        if len(args) > 1:
            print(self.executor.execute_line(args))
        elif len(args) == 1:
            print(args[0])
        else:
            print("DEBUG")

    def _obj(self, *args):
        if len(args) == 1:
            return args[0]
        return args



from typing import Tuple, Optional, Pattern, Dict, List
import regex as re
from .error import ArgumentNotExistError
import logging
import json


class ArgsResolver:
    """
    Resolve the arguments in the script
    """
    logging = logging.getLogger("lineup_lang")

    def resolve(self, script: str, **kwargs: str) -> str:
        regex: List[Tuple[Pattern, int, int | None, Optional[str]]] = [
            (r"\$\{([\w\-\_]+)(:(.+?))?\}", 1, 3),
            (r"\$([\w\-\_]+)", 1, None),
            (r"\$\((\w+):(.+?)\)", 1, 2, "{0} is deprecated")
        ]
        for i, r in enumerate(regex):
            matches = re.finditer(r[0], script)
            for match in matches:
                self.logging.debug(f"Match: {match.group(0)} with regex: {i} ({r[0]})")
                if len(r) == 4:
                    self.logging.warning(r[3].format(match.group(0)))
                script = self._modify_script(script, match, r, kwargs)
        return script

    def _modify_script(self, script: str, match: re.Match, regex: Tuple[Pattern, int, int | None], kwargs: Dict[str, str]) -> str:
        """
        Modify the script with the arguments
        """
        keyname = match.group(regex[1])
        default_value = None
        if regex[2]:
            default_value = match.group(regex[2])
        value = kwargs.get(keyname, default_value)
        if value is None:
            raise ArgumentNotExistError(f"'{keyname}' not exist in '{kwargs}'")
        value = json.dumps(value)
        value = value.replace("#", "\\#")
        return script.replace(match.group(0), value)
